#
# Copyright(c) 2006-2020 ThreatSTOP, Inc. All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property
# of ThreatSTOP, Inc. and its suppliers, if any. The intellectual and
# technical concepts contained herein are proprietary to ThreatSTOP, Inc.
# and its suppliers and certain aspects thereof are protected by United
# States Patent No. 8,533,822 and United States Patent No. 8,869,237,
# and are protected by trade secret or copyright law.
#
# US Government Entities: The ThreatSTOP service, software and
# documentation, as applicable, are "commercial computer software" and
# "commercial computer software documentation" developed exclusively at
# private expense by Threatstop, Inc. ("Threatstop"). Pursuant to FAR
# 12.212 or DFARS 227 7202 and their successors, as applicable, use,
# reproduction and disclosure of the software and documentation is
# governed by the terms of Threatstop's commercial agreement.
#

# file : ts-policy.sh
# This module supports policy lookup and generation.  It depends
# upon ts-dns.sh for actual DNS queries, as well as ts-common.sh for many
# things.

# genPolicyList() - generate a policy file, based on passed parameters
#
# args :
# - zone name (FQDN)
# - full path to file in which the list will be placed
# - optional sed arguments, that can delete lines (like '/foo/d;')
#
# returns :
# 0 for success
# 1 for failure
genPolicyList() {

    local type=$1
    local zone=$2
    local target_file=$3
    local sed_filter="$4"
    local ptr
    local apl
    local apls
    local file
    local rval=0

    # to get all CIDRs, we always
    # "look up APL records for all PTR records of the passed zone"
    echoMsg DEBUG "Generating $type list(s)"

    for file in ${target_file} ${target_file}.all; do

        if [ -s "${file}" ]; then

            mv ${file} ${file}.bak
        fi
    done

    # get PTRs
    performLookup ptr $zone "" "" "" 's/^[^ ]+ domain name pointer //p;' "true"
    if [ $? -eq 0 ]; then

        # get APLs for each PTR
        for ptr in ${LOOKUP_OUT}; do

            echoMsg DEBUG "Looking up APL records for PTR=$ptr"
            performLookup apl $ptr "" 's/^1:(.+)$/\1;/p;' "" 's/^[^ ]+ has APL record 1:(.+)$/\1;/p;' "true"

            if [ $? -eq 0 ]; then

                if [ "${LOOKUP_OUT}" != "" ]; then

                     # create file with all, *except*  anything filtered (can be empty string) by the user,
                     # and only the first MAXPOLICYSIZE lines
                     echo "${LOOKUP_OUT}" | sed ${SED_EXT_REGEXP} "${sed_filter}" >> ${target_file}.all
                fi
            else

                echoMsg ERROR "Encountered problems querying for APL records for $ptr"
                rval=1
            fi
        done

        if [ -s "${target_file}.all" ]; then
            all_size=`cat ${target_file}.all | wc -l`
            if [ ${all_size} -gt ${MAXPOLICYSIZE} ]; then
                echoMsg DISPLAY " ----------------------------------------------------------- "
                echoMsg INFO "  Maximum Policy Size ${MAXPOLICYSIZE} reached."
                echoMsg INFO "  All remaining addresses will be truncated"
                echoMsg DISPLAY " ----------------------------------------------------------- "
                echoMsg DEBUG "Policy ${target_file} had $all_size records, Maximum Policy Size ${MAXPOLICYSIZE} reached. Records will be truncated."
            fi

            echoMsg DEBUG "Creating ${target_file} with maximum size of ${MAXPOLICYSIZE}"
            cat "${target_file}.all" | sort -u | sed "$(expr ${MAXPOLICYSIZE} + 1),\$d;" > ${target_file}
        else

            echoMsg DEBUG "NO data was returned for the policy '$zone'"
            rval=1
        fi
        EFFECTIVE_CAP="$MAXPOLICYSIZE"
        if [ ${BLOCK_PREFIX_LIST_LIMIT} -gt 1 ] && [ $type = "block" ];then
            echoMsg INFO "Number of Prefix Lists available : ${BLOCK_PREFIX_LIST_LIMIT}"
            CALCULATED_CAP=$(expr $BLOCK_PER_PREFIX_LIST \* $BLOCK_PREFIX_LIST_LIMIT)
            echoMsg INFO "Prefix List Capactiy: ${BLOCK_PER_PREFIX_LIST}"
            echoMsg INFO "Combined Prefix List Calculated Capactiy: ${CALCULATED_CAP}"
            if [ ${MAXPOLICYSIZE} -gt ${CALCULATED_CAP} ];then
                EFFECTIVE_CAP="$CALCULATED_CAP"
            fi
        fi
        # BLOCK_PREFIX_LIST_LIMIT=10 # to test sample values
        # BLOCK_PER_PREFIX_LIST=1000 # defaults to 85k
        # if [ ${BLOCK_PREFIX_LIST_LIMIT} -gt 1 ] && [ ${all_size} -gt ${BLOCK_PER_PREFIX_LIST} ];then
        if [ ${BLOCK_PREFIX_LIST_LIMIT} -gt 1 ] && [ "[${type}]" = "[block]" ];then
            # fill to $BLOCK_PER_PREFIX_LIST then create new list.
            i=1
            startline=1
            endline=${BLOCK_PER_PREFIX_LIST}
            end=${BLOCK_PREFIX_LIST_LIMIT}
            total=$all_size
            echoMsg INFO "Initial Total: $total"
            if [ $all_size -gt $MAXPOLICYSIZE ];then
                echoMsg INFO "Max Policy Size Total: $MAXPOLICYSIZE"
                total=$MAXPOLICYSIZE
            fi
            addrs_in_file=`cat ${target_file}|wc -l`
            if [ $addrs_in_file -lt $all_size ];then
                total=$addrs_in_file
                echoMsg DEBUG "Addresses in file ${target_file} Total: $total"
            fi
            # echoMsg INFO "Number of Addresses in ${target_file}: $total"
            if [ $type = "block" ];then
                echoMsg DISPLAY " ----------------------------------------------------------- "
                echoMsg INFO "  Total Prefix List(s) Effective Combined Capactiy: ${EFFECTIVE_CAP}"
                echoMsg DISPLAY " ----------------------------------------------------------- "
            fi
            # cleanup previous lists
            if [ -f ${target_file}.$i ];then
                echoMsg DEBUG "Cleanup previous prefix lists ${target_file}.*"
                rm -f "${target_file}.*"
            fi
            while [ $i -le $end ]; do
                if [ $startline -gt $total ];then
                    echoMsg DEBUG "no more prefixes to add..."
                    echoMsg DEBUG "clearing out remaining list ${target_file}.$i"
                    cat /dev/null > "${target_file}.$i"
                    i=$(($i+1))
                    continue
                fi
                echoMsg INFO "Building list ${target_file}.$i $startline -> $endline";
                cat $target_file | sed -n "${startline},${endline}p" > "${target_file}.$i"
                startline=$(($startline + ${BLOCK_PER_PREFIX_LIST}))
                endline=$(expr $startline + ${BLOCK_PER_PREFIX_LIST})
                endline=$(expr $endline - 1)
                i=$(($i+1))
            done

            # echo "STOP HERE FOR DEV";
            # exit 2;


        fi

    else

        echoMsg ERROR "Encountered problems querying for PTR records for $zone"
        rval=1
    fi

    return $rval
}

# need to ensure config sanity prior to grabbing allow & block lists
# list names, disk space, allowed to query ThreatSTOP servers
#
# returns :
# 0 for success
# 1 for failure
preCheck() {

    local rval=0
    local meg_check_out
    local meg_check_megs
    local meg_check_partition

    if [ "${BLOCK_LIST}" = "" ]; then

        echoMsg ERROR "Block list not defined"
        rval=1
    fi
    if [ "${ALLOW_LIST}" = "" ]; then

        echoMsg ERROR "Allow list not defined"
        rval=1
    fi

    echoMsg DEBUG "checking for adequate disk space"
    meg_check_out=$(df -P -m ${INSTALL_DIR})
    meg_check_megs=$(echo "${meg_check_out}" | sed ${SED_EXT_REGEXP} '/^Filesystem/d; s/^\/[-_/0-9a-z]+[ ]+([0-9]+).*$/\1/;')
    meg_check_partition=$(echo "${meg_check_out}" | sed ${SED_EXT_REGEXP} '/^Filesystem/d; s/^(\/[-_/0-9a-z]+).*$/\1/;')
    echo "${meg_check_megs}" | grep -q -E '^[0-9]+$'
    if [ $? -eq 0 ]; then

        echoMsg DEBUG "checking ${meg_check_megs} < ${FREE_DISK_SPACE_MEGS}"
        if [ ${meg_check_megs} -lt ${FREE_DISK_SPACE_MEGS} ]; then

            echoMsg ERROR "Only ${meg_check_megs} MBs available on ${meg_check_partition} partition; ${FREE_DISK_SPACE_MEGS} free MBs required"
            echoMsg ERROR "The ThreatSTOP list update will not run"
            rval=1
        else

            echoMsg DEBUG "disk space is okay"
        fi
    else

        echoMsg DEBUG "Couldn't determine free megabytes; continuing anyway"
    fi

    return $rval
}

# make sure the generated lists are sane
# (of correct format and reasonable size)
#
# globals :
# - ALLOW_LIST_PATH
# - BLOCK_LIST_PATH
#
# returns :
# 0 for success
# 1 for failure
postCheck() {

    local rval=0
    local allow_lines
    local block_lines
    local valid_allow_cidr_lines
    local valid_block_cidr_lines
    local rogue_lines
    local cidr_sed='/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\/[0-9]+;$/p;'

    if [ ! -e "${ALLOW_LIST_PATH}" -o ! -e "${BLOCK_LIST_PATH}" ]; then

        echoMsg ERROR "One or both policy lists were not generated"
        return 1
    fi

    allow_lines=$(wc -l ${ALLOW_LIST_PATH} | sed -n ${SED_EXT_REGEXP} 's/^ *([0-9]+) *.+$/\1/p;')
    block_lines=$(wc -l ${BLOCK_LIST_PATH} | sed -n ${SED_EXT_REGEXP} 's/^ *([0-9]+) *.+$/\1/p;')
    valid_allow_cidr_lines=$(cat ${ALLOW_LIST_PATH} | sed -n ${SED_EXT_REGEXP} "${cidr_sed}" | wc -l | sed -n ${SED_EXT_REGEXP} 's/^ *([0-9]+)$/\1/p;')
    valid_block_cidr_lines=$(cat ${BLOCK_LIST_PATH} | sed -n ${SED_EXT_REGEXP} "${cidr_sed}" | wc -l | sed -n ${SED_EXT_REGEXP} 's/^ *([0-9]+)$/\1/p;')

    echoMsg DEBUG "allow lines : total=${allow_lines}, valid=${valid_allow_cidr_lines}"
    echoMsg DEBUG "block lines : total=${block_lines}, valid=${valid_block_cidr_lines}"

    if [ "${valid_allow_cidr_lines}" -eq 0 -o "${valid_allow_cidr_lines}" -ne "${allow_lines}" ]; then

        echoMsg ERROR "Generated allow list cannot be used; it has ${valid_allow_cidr_lines} out of ${allow_lines} valid lines"
        rval=1
    fi
    if [ "${valid_block_cidr_lines}" -eq 0 -o "${valid_block_cidr_lines}" -ne "${block_lines}" ]; then

        echoMsg ERROR "Generated block list cannot be used; it has ${valid_block_cidr_lines} out of ${block_lines} valid lines"
        rval=1
    fi

    # need to guard against dangerous CIDRs that can NEVER be loaded into block lists
    echoMsg INFO "Final sanity checks..."
    rogue_lines=$(cat ${BLOCK_LIST_PATH} | sed -n ${SED_EXT_REGEXP} 's/^(0\.0\.0\.0\/0);$/\1/p;') # cannot have 0.0.0.0/0
    if [ "x${rogue_lines}" != "x" ]; then

        echoMsg ERROR "Detected malicious entries in the block list"
        echoMsg ERROR "${rogue_lines}"
        rval=1
    fi

    return $rval
}
